setClass("SOAPClientInterface",
           representation(functions = "list",
                          classes = "list"))



capitalizeFirstLetter =
function(word)
{
  els = strsplit(word, "")[[1]]
  els[1] = toupper(els[1])
  paste(els, collapse="")
}

convertToSName =
function(name, useDash = ifelse("UseDashInSOAPNames" %in% names(options()), getOption("UseDashInSOAPNames"), TRUE))
{
  if(useDash)
    return(name)
  
  els = strsplit(name, "_")[[1]]
  if(length(els) == 1)
    return(els)
  
  els[2:length(els)] = sapply(els[-1], capitalizeFirstLetter)

  paste(els, collapse="")
}



genSOAPClientInterface =
  #
  # def is from reading the WSDL file
  # name is the prefix for the methods
  # env is where to put the methods.
  #
function(operations = def@operations[[1]], def, name, env = new.env(), where = 1, verbose = FALSE)  
{
 funs = list()

 for(i in operations) {
   if(verbose)
     cat("Operation", i@name, "\n")

   funs[[convertToSName(i@name)]] = createOperationDefinition(i, def@server, def@types, env)
 }


 if(!is.null(where)) {
  classes = list()
  for(i in def@types) {
   if(is(i$definition, "ClassDefinition")) {
     if(verbose)
       cat("Defining class for", i$name,"\n")
     classes[[i$name]] = createTypeClass(i, def@types, where = where)
   }
  }
 } else
   classes = NULL

 ans = new("SOAPClientInterface", functions = funs, classes =classes)

 ans
}


createOperationDefinition =
function(.operation, .defaultServer, .types, env = globalenv())
{
   # resolve the parameter types
 .operation@parameters = lapply(.operation@parameters, resolve, .types)

 .operation@returnValue = resolve(.operation@returnValue, .types)

 
 if(.operation@action == "")
     # NULL is not used, so just NULL for placeholder.
   .operation@action[1] = .SOAPAction(.operation@action, .operation@name, NULL, .operation@namespace)

 dotArgs = ""
 if(length(.operation@parameters))
   dotArgs = paste(paste(paste("'", names(.operation@parameters), "'", sep=""),
                                       sapply(names(.operation@parameters), convertToSName), sep = " = "),
                   collapse=", ")  
 txt = paste("function(",
              paste(c(sapply(names(.operation@parameters), convertToSName), "server = .defaultServer, .convert = .operation@returnValue, .opts = list()"), collapse=",  "),
             ")\n{\n",
             "  .SOAP(server, .operation@name, ",
                           dotArgs,
                           ifelse(length(.operation@parameters), ", ", ""),
                           "action = .operation@action, ",
                           "xmlns = .operation@namespace, ",
                           ".types = .operation@parameters, ",
                           ".convert = .convert,",
                           ".opts = .opts) ","\n",
#                   "if(.debug) return(val)\n  convertFromSOAP(val, .operation@returnValue)",
                  "\n}", sep="\n\t")  
 f = eval(parse(text = txt))

 attr(f, "source") <- txt
 
 rm(txt)
 rm(dotArgs)
 
 f
}


setGeneric("convertFromSOAP", function(val, type) standardGeneric("convertFromSOAP"))

setMethod("convertFromSOAP", "SOAPResult", 
          function(val, type) {
            node <- parseSOAP(val@content, asText = TRUE)
            node <- getReturnNode(node)
            
            fromSOAP(node, type = type)
          })



createTypeClass =
function(type, types, where = globalenv(), parentClass = "VirtualSOAPClass")
{
 repn = createClassRepresentation(type$definition, types)
 setClass(type$name, representation = repn, where = where, contains = parentClass)
}


newSOAPClass =
function(node, className, converters = SOAPPrimitiveConverters, type = NULL)
{
 obj = new(className)
 
 for(i in slotNames(className)) {
   tmp = node[[i]]
   if(!is.null(type))
     slotType = type@slotTypes[[i]]
   else
     slotType = NULL

   slot(obj, i) <- fromSOAP(tmp, type = slotType, findReturn = FALSE) #XXX Need SOAP type here!
 }

  obj
}


createClassRepresentation =
function(type, types)
{
 repn = list()
 for(i in names(type@slotTypes)) {
   repn[[i]] = mapSOAPTypeToS(type@slotTypes[[i]])
 }

 repn
}

mapSOAPTypeToS =
function(type)
{
    which = sapply(SOAPTypes, function(x) type == x)
    if(!all(is.na(which)))
      return(names(SOAPTypes)[which][1])

    stop("Can't match SOAP type ", type)
}





if(FALSE) {

# f = function(){}
# formals(f) <- alist(server=, kid=, threshold=, orgs=)
  
 body(f) <- substitute({
  val = .SOAP(server,
              .opName,
              action = .action,
              xmlns = .namespace,
              .types = ..types)
  }, 
  list(.opName= operation@name,
       .action = operation@action,
       .namespace = operation@namespace,
       ..types = operation@parameters))

  # now put the arguments in.
 e = body(f)[[2]]
 kk = e[4:length(e)]
# for(i in )
 k = body(f)[[2]][[3]]
}

# body(f) = substitute(body(f), list(.opname=operation@name))
